from ..bindings import execute_graph


def test_upload_context(mock_icat_client):
    graph = {
        "graph": {"id": "demo", "label": "demo", "schema_version": "1.1"},
        "links": [],
        "nodes": [
            {
                "default_inputs": [
                    {"name": "a", "value": 42},
                    {"name": "delay", "value": 0},
                ],
                "id": "task0",
                "task_identifier": "ewokscore.tests.examples.tasks.sumtask.SumTask",
                "task_type": "class",
            },
        ],
    }

    execute_graph(
        graph,
        inputs=[{"name": "delay", "value": 0, "all": False}],
        upload_parameters={},
    )

    args, kwargs = mock_icat_client.return_value.store_processed_data.call_args

    assert not args
    assert set(kwargs) == {"metadata"}
    assert set(kwargs["metadata"]) == {"startDate", "endDate"}
    duration = kwargs["metadata"]["endDate"] - kwargs["metadata"]["startDate"]

    delay = 1
    margin = 0.5

    execute_graph(
        graph,
        inputs=[{"name": "delay", "value": delay + margin, "all": False}],
        upload_parameters={},
    )

    args, kwargs = mock_icat_client.return_value.store_processed_data.call_args

    assert not args
    assert set(kwargs) == {"metadata"}
    assert set(kwargs["metadata"]) == {"startDate", "endDate"}
    duration_with_delay = (
        kwargs["metadata"]["endDate"] - kwargs["metadata"]["startDate"]
    )

    difference = duration_with_delay - duration
    assert difference.total_seconds() > delay



import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


@contextmanager
def notebook_path(*args) -> Generator[Path, None, None]:
    source = importlib_resources.files(__name__).joinpath(*args)
    with importlib_resources.as_file(source) as path:
        if not path.is_file():
            raise FileNotFoundError(f"Not a file: '{source}'")
        yield path



import pytest

from .._requirements.pip.sanitize import sanitize_requirements


def test_normal_requirement():
    req = ["ewoks==1.1.0"]
    sanitized, warnings = sanitize_requirements(req)
    assert sanitized == ["ewoks==1.1.0"]
    assert warnings == []


def test_editable_ssh_vcs_url_normalized():
    ssh_project_name = "querypool"
    ssh_project_url = "gitlab.esrf.fr/dau/querypool.git"
    ssh_project_commit = "ab6acc7e140ed33eb896b1336a5a5aac6b60cc0f"
    req = [
        f"-e git+ssh://git@{ssh_project_url}@{ssh_project_commit}#egg={ssh_project_name}"
    ]

    sanitized, warnings = sanitize_requirements(req)

    expected_sanitized = [
        f"{ssh_project_name} @ git+https://{ssh_project_url}@{ssh_project_commit}"
    ]
    expected_warnings = [
        f"Normalize VCS requirement 'git+ssh://git@{ssh_project_url}@{ssh_project_commit}#egg={ssh_project_name}' "
        f"to '{ssh_project_name} @ git+ssh://git@{ssh_project_url}@{ssh_project_commit}'",
        f"Normalize requirement '{ssh_project_name} @ git+ssh://git@{ssh_project_url}@{ssh_project_commit}' "
        f"to '{ssh_project_name} @ git+https://{ssh_project_url}@{ssh_project_commit}'",
    ]

    assert sanitized == expected_sanitized
    assert warnings == expected_warnings


@pytest.mark.parametrize("exists", [True, False])
def test_editable_local_path_with_comment_replacement(tmp_path, exists):
    path = tmp_path / "repo_name"
    if exists:
        path.mkdir()

    comment = "# Editable Git install with no remote (project_name==1.0.0)"
    replacement = "project_name==1.0.0"
    req = [comment, f"-e {path}"]

    sanitized, warnings = sanitize_requirements(req)

    assert sanitized == [replacement]
    assert warnings == [f"Replaced editable install '{path}' with '{replacement}'."]


@pytest.mark.parametrize("exists", [True, False])
def test_editable_local_path_without_comment_replacement(tmp_path, exists):
    path = tmp_path / "repo_name"
    if exists:
        path.mkdir()
        warning = f"Editable path exists locally: '{path}'"
    else:
        warning = f"Editable path does not exist locally: '{path}'"

    req = [f"-e {path}"]

    sanitized, warnings = sanitize_requirements(req)

    assert sanitized == [f"-e {path}"]
    if exists:
        assert warnings == [warning]
    else:
        assert warnings == [warning]


def test_branch_specified_requirement():
    project_name = "ewoksutils"
    project_url = "gitlab.esrf.fr/workflow/ewoks/ewoksutils.git"
    project_branch = "main"

    req = [f"{project_name}@ git+https://{project_url}@{project_branch}"]

    sanitized, warnings = sanitize_requirements(req)

    # No warnings expected here (assuming valid format)
    assert sanitized == [f"{project_name}@ git+https://{project_url}@{project_branch}"]
    assert warnings == []


def test_invalid_requirement_warning():
    project_url = "gitlab.esrf.fr/workflow/ewoks/ewoksutils.git"

    req = [f"git+https://{project_url}"]

    sanitized, warnings = sanitize_requirements(req)

    assert sanitized == [f"git+https://{project_url}"]
    assert any("Possibly invalid requirement format" in w for w in warnings)



import datetime
import logging
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

from ewokscore.events.contexts import RawExecInfoType
from ewokscore.events.contexts import job_context
from ewokscore.graph import TaskGraph
from ewokscore.graph.inputs import graph_inputs_as_table
from tabulate import tabulate

from . import _engines
from . import graph_cache
from ._requirements.pip.extract import add_current_env_pip_requirements
from ._requirements.pip.extract import extract_pip_requirements
from ._requirements.pip.install import pip_install
from .errors import AbortException

try:
    from ewoksjob.bindings import submit_graph as _submit_graph
    from ewoksjob.client.futures import FutureInterface
except ImportError:
    _submit_graph = None
    FutureInterface = Any

try:
    from pyicat_plus.client import defaults as icat_defaults
    from pyicat_plus.client.main import IcatClient
except ImportError:
    IcatClient = None
    icat_defaults = None


__all__ = ["execute_graph", "load_graph", "save_graph", "convert_graph", "submit_graph"]

logger = logging.getLogger(__name__)


def execute_graph(
    graph,
    engine: Optional[str] = None,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    varinfo: Optional[dict] = None,
    execinfo: RawExecInfoType = None,
    task_options: Optional[dict] = None,
    outputs: Optional[List[dict]] = None,
    merge_outputs: Optional[bool] = True,
    environment: Optional[dict] = None,
    convert_destination: Optional[Any] = None,
    save_options: Optional[dict] = None,
    upload_parameters: Optional[dict] = None,
    **execute_options,
) -> Optional[dict]:
    with job_context(execinfo, engine=engine) as execinfo:
        with _upload_context(upload_parameters):
            if environment:
                environment = {k: str(v) for k, v in environment.items()}
                os.environ.update(environment)

            # Load the graph
            if load_options is None:
                load_options = dict()
            graph = load_graph(graph, inputs=inputs, **load_options)

            # Save the graph (with inputs)
            if convert_destination is not None:
                convert_graph(graph, convert_destination, save_options=save_options)

            # Execute the graph
            engine_api = _engines.get_execution_engine(engine)
            result = engine_api.execute_graph(
                graph,
                varinfo=varinfo,
                execinfo=execinfo,
                task_options=task_options,
                outputs=outputs,
                merge_outputs=merge_outputs,
                **execute_options,
            )

            return result


@contextmanager
def _upload_context(upload_parameters: Optional[dict]) -> Generator[None, None, None]:
    if upload_parameters is None:
        yield
        return

    if IcatClient is None:
        raise RuntimeError("requires the 'pyicat-plus' package")

    metadata = upload_parameters.setdefault("metadata", {})
    if "startDate" not in metadata:
        metadata["startDate"] = datetime.datetime.now().astimezone()

    yield

    # Only upload when no exception is raised.

    if "endDate" not in metadata:
        metadata["endDate"] = datetime.datetime.now().astimezone()

    metadata_urls = upload_parameters.pop(
        "metadata_urls", icat_defaults.METADATA_BROKERS
    )
    client = IcatClient(metadata_urls=metadata_urls)
    logger.info(
        "Upload processed dataset '%s' metadata to ICAT: %s",
        upload_parameters.get("dataset"),
        upload_parameters.get("path"),
    )
    client.store_processed_data(**upload_parameters)


def submit_graph(
    graph,
    _celery_options: Optional[dict] = None,
    resolve_graph_remotely: Optional[bool] = None,
    load_options: Optional[dict] = None,
    **options,
) -> FutureInterface:
    """Submit a workflow to be executed remotely. The workflow is
    resolved on the client-side by default (e.g. load from a file)
    but can optionally be resolved remotely.
    """
    if _submit_graph is None:
        raise RuntimeError("requires the 'ewoksjob' package")
    return _submit_graph(
        graph,
        _convert_graph=_load_graph,
        _celery_options=_celery_options,
        resolve_graph_remotely=resolve_graph_remotely,
        load_options=load_options,
        **options,
    )


def _load_graph(graph, load_options: Optional[dict] = None) -> dict:
    return convert_graph(
        graph, None, load_options=load_options, save_requirements=False
    )


@graph_cache.cache
def load_graph(
    graph: Any,
    inputs: Optional[List[dict]] = None,
    representation: Optional[str] = None,
    root_dir: Optional[Union[str, Path]] = None,
    root_module: Optional[str] = None,
    **load_options,
) -> TaskGraph:
    """When load option `graph_cache_max_size > 0` is provided, the graph will cached in memory.
    When the graph comes from external storage (for example a file) any changes
    to the external graph will require flushing the cache with `graph_cache_max_size = 0`.
    """
    engine_api, representation = _engines.get_serialization_engine(
        graph, representation=representation
    )
    return engine_api.deserialize_graph(
        graph,
        inputs=inputs,
        representation=representation,
        root_dir=root_dir,
        root_module=root_module,
        **load_options,
    )


def save_graph(
    graph: TaskGraph,
    destination,
    representation: Optional[str] = None,
    **save_options,
) -> Union[str, dict]:
    engine_api, representation = _engines.get_serialization_engine(
        destination, representation=representation
    )
    return engine_api.serialize_graph(
        graph, destination, representation=representation, **save_options
    )


def convert_graph(
    source,
    destination,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    save_options: Optional[dict] = None,
    save_requirements: bool = True,
) -> Union[str, dict]:
    if load_options is None:
        load_options = dict()
    if save_options is None:
        save_options = dict()
    graph = load_graph(source, inputs=inputs, **load_options)
    if save_requirements:
        graph = add_current_env_pip_requirements(graph)
    return save_graph(graph, destination, **save_options)


def show_graph(
    source,
    inputs: Optional[List[dict]] = None,
    load_options: Optional[dict] = None,
    original_source: Optional[str] = None,
    column_widths: Optional[Dict[str, Optional[int]]] = None,
) -> Union[str, dict]:
    if load_options is None:
        load_options = dict()
    graph = load_graph(source, inputs=inputs, **load_options)
    _print_graph(graph, column_widths=column_widths, original_source=original_source)


def _print_graph(
    graph: TaskGraph,
    column_widths: Optional[Dict[str, Optional[int]]] = None,
    original_source: Optional[str] = None,
) -> None:
    column_names, rows, metadata, footnotes = graph_inputs_as_table(
        graph, column_widths=column_widths
    )
    print()
    if original_source:
        print(f"Workflow: {original_source}")
    else:
        print("Workflow:")
    for key, value in metadata.items():
        label = key.replace("_", " ").capitalize()
        print(f"{label}: {value}")
    if rows:
        print(tabulate(rows, headers=column_names, tablefmt="fancy_grid"))
    else:
        print("No workflow inputs parameters detected!")
    for footnote in footnotes:
        print(footnote)


def install_graph(
    source,
    skip_prompt: bool = False,
    python_path: Optional[str] = None,
    load_options: Optional[dict] = None,
):
    if load_options is None:
        load_options = dict()
    graph = load_graph(source, **load_options)

    requirements = graph.requirements
    if requirements is None:
        logger.warning(
            "Requirements field is empty. Trying to extract requirements automatically..."
        )
        requirements = extract_pip_requirements(graph)
        logger.info(f"Extracted the following requirements: {requirements}")

    if python_path is None:
        python_path = sys.executable

    if skip_prompt:
        pip_install(requirements, python_path)
        return

    requirements_as_str = "\n".join(requirements)

    answer = input(
        f"{requirements_as_str}\nThis will install the above packages via {python_path} -m pip install. Do you want to proceed (y/N)?"
    )
    if answer.lower() == "y" or answer.lower() == "yes":
        pip_install(requirements, python_path)
    else:
        raise AbortException()



from ewoksutils.tests.conftest import cli_interface  # noqa F401
from ewoksutils.tests.conftest import graph_directory  # noqa F401

from ..cli_utils import cli_convert_utils


def test_cli_convert(cli_interface):  # noqa F811
    argv = [
        "acyclic1",
        "test.json",
        "--test",
        "-p",
        "a=1",
        "-p",
        "task1:b=test",
        "--src-format",
        "yaml",
        "--dst-format",
        "json",
    ]
    cli_args = cli_interface(
        argv,
        cli_convert_utils.convert_arguments,
        cli_convert_utils.parse_convert_arguments,
    )

    assert list(cli_args.graphs) == ["acyclic1"]

    convert_options = {
        "inputs": [
            {"all": False, "name": "a", "value": 1},
            {"id": "task1", "name": "b", "value": "test"},
        ],
        "load_options": {"representation": "test_core"},
        "save_options": {"representation": "json"},
    }
    assert cli_args.convert_options == convert_options

    argv = ["acyclic1", "test.json"]
    cli_args = cli_interface(
        argv,
        cli_convert_utils.convert_arguments,
        cli_convert_utils.parse_convert_arguments,
    )

    assert cli_args.destinations == ["test.json"]

    argv = ["acyclic1", ".json"]
    cli_args = cli_interface(
        argv,
        cli_convert_utils.convert_arguments,
        cli_convert_utils.parse_convert_arguments,
    )

    assert cli_args.destinations == ["acyclic1.json"]

    argv = ["acyclic1", "json"]
    cli_args = cli_interface(
        argv,
        cli_convert_utils.convert_arguments,
        cli_convert_utils.parse_convert_arguments,
    )

    assert cli_args.destinations == ["acyclic1.json"]


def test_cli_convert_search(cli_interface, tmp_path, graph_directory):  # noqa F811
    argv = [
        str(tmp_path / "subdir" / "*.json"),
        str(tmp_path / "*.json"),
        "_suffix.json",
        "--search",
    ]
    cli_args = cli_interface(
        argv,
        cli_convert_utils.convert_arguments,
        cli_convert_utils.parse_convert_arguments,
    )

    assert len(cli_args.graphs) == 22
    assert cli_args.graphs == graph_directory



import sys
import traceback
from argparse import ArgumentDefaultsHelpFormatter
from argparse import ArgumentParser
from argparse import Namespace
from pprint import pprint
from subprocess import CalledProcessError
from typing import Any
from typing import List
from typing import Literal
from typing import Optional
from typing import Union

from ewoksutils.cli_utils import cli_cancel_utils
from ewoksutils.cli_utils import cli_execute_utils
from ewoksutils.cli_utils import cli_submit_utils
from ewoksutils.cli_utils.cli_argparse import add_to_parser

from .bindings import _load_graph
from .bindings import convert_graph
from .bindings import execute_graph
from .bindings import install_graph
from .bindings import show_graph
from .cli_utils import cli_convert_utils
from .cli_utils import cli_install_utils
from .cli_utils import cli_show_utils
from .errors import AbortException

try:
    from ewoksjob.cli_utils.cancel import command_cancel as _command_cancel
    from ewoksjob.cli_utils.submit import command_submit as _command_submit
except ImportError:
    _command_submit = None
    _command_cancel = None


def create_argument_parser(shell: bool = False) -> ArgumentParser:
    parser = ArgumentParser(
        description="Extensible WOrKflow System CLI",
        prog="ewoks",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    subparsers = parser.add_subparsers(help="Commands", dest="command")
    execute = subparsers.add_parser(
        "execute",
        help="Execute a workflow",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    submit = subparsers.add_parser(
        "submit",
        help="Schedule a workflow execution",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    cancel = subparsers.add_parser(
        "cancel",
        help="Abort a scheduled workflow execution",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    convert = subparsers.add_parser(
        "convert",
        help="Convert a workflow",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    install = subparsers.add_parser(
        "install",
        help="Install requirements of a workflow",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    show = subparsers.add_parser(
        "show",
        help="Show workflow information",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    add_to_parser(execute, cli_execute_utils.execute_arguments(shell=shell))
    add_to_parser(submit, cli_submit_utils.submit_arguments(shell=shell))
    add_to_parser(cancel, cli_cancel_utils.cancel_arguments(shell=shell))
    add_to_parser(convert, cli_convert_utils.convert_arguments(shell=shell))
    add_to_parser(install, cli_install_utils.install_arguments(shell=shell))
    add_to_parser(show, cli_show_utils.show_arguments(shell=shell))
    return parser


def command_execute(
    cli_args: Namespace, shell: bool = False
) -> Union[List[dict], Literal[0, 1]]:
    cli_execute_utils.parse_execute_argument(cli_args, shell=shell)

    return_code = 0
    keep_results = []
    for workflow, graph in zip(cli_args.workflows, cli_args.graphs):
        print("###################################")
        print(f"# Execute workflow '{workflow}'")
        print("###################################")
        try:
            results = execute_graph(
                graph, engine=cli_args.engine, **cli_args.execute_options
            )
        except Exception as ex:
            traceback.print_exc()
            print("FAILED")
            results = ex
            return_code = 1
        else:
            if cli_args.outputs == "none":
                if results is None:
                    print("FAILED")
                else:
                    print("FINISHED")
            else:
                print("")
                print("RESULTS:")
                pprint(results)
                print("")
                print("FINISHED")
            if results is None:
                return_code = 1
        finally:
            print()
        if not shell:
            keep_results.append(results)

    if shell:
        return return_code
    return keep_results


def command_submit(
    cli_args: Namespace, shell: bool = False
) -> Union[List[dict], Literal[0, 1]]:
    if _command_submit is None:
        raise RuntimeError("requires the 'ewoksjob>=1.3' package")
    return _command_submit(cli_args, _convert_graph=_load_graph, shell=shell)


def command_cancel(cli_args: Namespace, shell: bool = False) -> Optional[Literal[0, 1]]:
    if _command_cancel is None:
        raise RuntimeError("requires the 'ewoksjob>=1.3' package")
    return _command_cancel(cli_args, shell)


def command_convert(
    cli_args: Namespace, shell: bool = False
) -> Optional[Literal[0, 1]]:
    cli_convert_utils.parse_convert_arguments(cli_args, shell=shell)
    for workflow, graph, destination in zip(
        cli_args.workflows, cli_args.graphs, cli_args.destinations
    ):
        convert_graph(graph, destination, **cli_args.convert_options)
        print(f"Converted {workflow} -> {destination}")
    if shell:
        return 0
    return None


def command_install(
    cli_args: Namespace, shell: bool = False
) -> Optional[Literal[0, 1]]:
    cli_install_utils.parse_install_arguments(cli_args, shell=shell)
    for workflow, graph in zip(cli_args.workflows, cli_args.graphs):
        try:
            install_graph(graph, cli_args.yes, cli_args.python)
        except CalledProcessError as e:
            print(f"Install failed for {workflow}: {e}")
        except AbortException:
            print(f"Install aborted for {workflow}")
        else:
            print(f"Installed requirements for {workflow}")
    if shell:
        return 0
    return None


def command_show(cli_args: Namespace, shell: bool = False) -> Optional[Literal[0, 1]]:
    cli_show_utils.parse_show_arguments(cli_args, shell=shell)
    for workflow, graph in zip(cli_args.workflows, cli_args.graphs):
        show_graph(graph, original_source=workflow, **cli_args.show_options)
    if shell:
        return 0
    return None


def command_default(
    cli_args: Namespace, shell: bool = False
) -> Optional[Literal[0, 1]]:
    if shell:
        return 1
    return None


def main(argv=None, shell: bool = True) -> Union[Any, Literal[0, 1]]:
    parser = create_argument_parser(shell=shell)

    if argv is None:
        argv = sys.argv
    cli_args = parser.parse_args(argv[1:])

    if cli_args.command == "execute":
        return command_execute(cli_args, shell=shell)
    elif cli_args.command == "submit":
        return command_submit(cli_args, shell=shell)
    elif cli_args.command == "cancel":
        return command_cancel(cli_args, shell=shell)
    elif cli_args.command == "convert":
        return command_convert(cli_args, shell=shell)
    elif cli_args.command == "install":
        return command_install(cli_args, shell=shell)
    elif cli_args.command == "show":
        return command_show(cli_args, shell=shell)
    else:
        parser.print_help()
        return command_default(cli_args, shell=shell)


if __name__ == "__main__":
    sys.exit(main())



import logging
from functools import lru_cache
from typing import Any
from typing import Callable
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple

from ewokscore.engine_interface import WorkflowEngine
from ewokscore.engine_interface import WorkflowEngineWithSerialization
from ewokscore.entry_points import EntryPoint
from ewokscore.entry_points import entry_points

_logger = logging.getLogger(__name__)


@lru_cache(1)
def get_engine_names() -> List[str]:
    return [name for name, _ in _iter_engine_class_loaders_with_name()]


@lru_cache(1)
def get_graph_representations() -> List[str]:
    return [
        representation
        for representation, _ in _iter_engine_class_loaders_with_representation()
    ]


def get_execution_engine(engine_name: Optional[str]) -> WorkflowEngine:
    if not engine_name or engine_name.lower() == "none":
        engine_name = "core"

    for name, load_engine_cls in _iter_engine_class_loaders_with_name():
        if name == engine_name:
            try:
                engine_cls = load_engine_cls()
            except Exception as e:
                _logger.warning(
                    f"Unable to properly load '{name}' engine. Error is {e}"
                )
                continue
            else:
                return engine_cls()

    raise RuntimeError(f"No engine found for graph execution: '{engine_name}'")


def get_serialization_engine(
    graph: Any, representation: Optional[str] = None
) -> Tuple[WorkflowEngine, str]:
    core_representation = representation
    if representation:
        for (
            representations,
            load_engine_cls,
        ) in _iter_engine_class_loaders_with_representation():
            if representation in representations:
                engine_cls = load_engine_cls()
                engine = engine_cls()
                return engine, representation
    else:
        for name, load_engine_cls in _iter_engine_class_loaders_with_name():
            try:
                engine_cls = load_engine_cls()
            except Exception as e:
                _logger.warning(
                    f"Unable to properly load serialization engine '{name}'. Error is {e}"
                )
                continue

            if not issubclass(engine_cls, WorkflowEngineWithSerialization):
                continue

            engine = engine_cls()
            representation = engine.get_graph_representation(graph)
            if representation:
                return engine, representation

    return get_execution_engine("core"), core_representation


def _iter_engine_class_loaders_with_name() -> (
    Generator[Tuple[str, Callable[[], None]], None, None]
):
    try:
        eps = entry_points(group="ewoks.engines")
    except Exception:
        return

    eps = sorted(eps, key=_sort_ewoks_engines)
    for ep in eps:
        yield ep.name, ep.load


def _iter_engine_class_loaders_with_representation() -> (
    Generator[Tuple[str, Callable[[], None]], None, None]
):
    try:
        eps = entry_points(group="ewoks.engines.serialization.representations")
    except Exception:
        return

    eps = sorted(eps, key=_sort_ewoks_engines)
    for ep in eps:
        yield ep.name, ep.load


def _sort_ewoks_engines(item: EntryPoint):
    return (item.name != "core", item)



import logging
import subprocess
from typing import Sequence

from .sanitize import sanitize_requirements

logger = logging.getLogger(__name__)


def pip_install(requirements: Sequence[str], python_path: str) -> int:
    requirements, warnings = sanitize_requirements(requirements)
    for warning in warnings:
        logger.warning(warning)
    # https://pip.pypa.io/en/stable/user_guide/#using-pip-from-your-program
    return subprocess.check_call([python_path, "-m", "pip", "install", *requirements])



import os
from typing import List


def parse_destinations(args) -> List[str]:
    dest_dirname = os.path.dirname(args.destination)
    basename = os.path.basename(args.destination)
    dest_basename, dest_ext = os.path.splitext(basename)
    if not dest_ext:
        dest_ext = dest_basename
        dest_basename = ""
        if not dest_ext.startswith("."):
            dest_ext = f".{dest_ext}"

    if len(args.workflows) == 1 and dest_basename:
        return [os.path.join(dest_dirname, f"{dest_basename}{dest_ext}")]

    destinations = list()
    for workflow in args.workflows:
        basename, _ = os.path.splitext(os.path.basename(workflow))
        destination = os.path.join(dest_dirname, f"{basename}{dest_basename}{dest_ext}")
        destinations.append(destination)

    return destinations



import os
import re
from typing import List
from typing import Sequence
from typing import Tuple
from urllib.parse import ParseResult
from urllib.parse import parse_qs
from urllib.parse import urlparse
from urllib.parse import urlunparse

from packaging.requirements import InvalidRequirement
from packaging.requirements import Requirement


def sanitize_requirements(requirements: Sequence[str]) -> Tuple[List[str], List[str]]:
    """Sanitize a list of requirements coming from 'pip freeze'.
    Returns a sanitized with warnings regarding applied changes.
    """
    sanitized = []
    warnings = []
    previous_comment = ""

    for requirement in requirements:
        requirement = requirement.strip()
        if not requirement:
            continue
        if requirement.startswith("#"):
            previous_comment = requirement
            continue

        editable = requirement.startswith("-e ")
        if editable:
            requirement = requirement[2:].strip()

        if editable and _is_path(requirement):
            requirement, warning = _handle_editable_path(requirement, previous_comment)
            if warning:
                warnings.append(warning)
            if _is_path(requirement):
                sanitized.append(f"-e {requirement}")
                continue

        requirement, warning = _normalize_vcs_url(requirement)
        if warning:
            warnings.append(warning)

        requirement, warning = _normalize_requirement(requirement)
        if warning:
            warnings.append(warning)

        sanitized.append(requirement)
        previous_comment = ""

    return sanitized, warnings


def _is_path(requirement: str) -> bool:
    """Return True if requirement looks like a local path (not a URL)."""
    return "://" not in requirement and ("/" in requirement or "\\" in requirement)


def _normalize_requirement(requirement: str) -> Tuple[str, str]:
    """Normalize the requirement or pass-through with warning when not valid."""
    try:
        req = Requirement(requirement)
    except InvalidRequirement:
        warning = f"Possibly invalid requirement format: {requirement!r}"
        return requirement, warning

    replacement = _format_requirement(req)
    if requirement.replace(" ", "") != replacement.replace(" ", ""):
        warning = f"Normalize requirement {requirement!r} to {replacement!r}"
        return replacement, warning

    return requirement, ""


def _format_requirement(req: Requirement) -> str:
    """Convert requirement to a string and replace ssh with https"""
    if req.url and req.url.startswith("git+ssh://"):
        https_url = re.sub(r"^git\+ssh://git@", "git+https://", req.url)
        return f"{req.name} @ {https_url}"
    return str(req)


def _normalize_vcs_url(requirement: str) -> Tuple[str, str]:
    """Normalize VCS URLs to the PEP 508 spec (name @ url)."""
    parsed = urlparse(requirement)

    if not parsed.fragment or "egg=" not in parsed.fragment:
        return requirement, ""

    qs = parse_qs(parsed.fragment)
    egg_names = qs.get("egg")
    if not egg_names:
        return requirement, ""

    egg_name = egg_names[0]
    base_url = _remove_url_fragment(parsed)
    new_requirement = f"{egg_name} @ {base_url}"
    warning = f"Normalize VCS requirement '{requirement}' to '{new_requirement}'"
    return new_requirement, warning


def _remove_url_fragment(parsed: ParseResult) -> str:
    return urlunparse(
        ParseResult(
            scheme=parsed.scheme,
            netloc=parsed.netloc,
            path=parsed.path,
            params=parsed.params,
            query=parsed.query,
            fragment="",
        )
    )


def _handle_editable_path(requirement: str, comment: str) -> Tuple[str, str]:
    """Replace editable path with comment substitution if available."""
    replacement = _requirement_from_comment(comment)
    if replacement:
        warning = f"Replaced editable install '{requirement}' with '{replacement}'."
        return replacement, warning
    if os.path.exists(requirement):
        warning = f"Editable path exists locally: '{requirement}'"
    else:
        warning = f"Editable path does not exist locally: '{requirement}'"
    return requirement, warning


_REQUIREMENT_COMMENT_RE = re.compile(r"\(([\w\.\-\[\]]+[=><!~]=[\w\.\-]+)\)")


def _requirement_from_comment(comment: str) -> str:
    """Extract requirement from a 'pip freeze' comment."""
    # When doing "pip freeze" in the same envrinment where "git status" gives this
    #   fatal: detected dubious ownership in repository at '/path/to/repo'
    # the requirements list will contain "-e /path/to/repo" with a warning that
    # contains a requirement message.
    match = _REQUIREMENT_COMMENT_RE.search(comment)
    if match:
        return match.group(1)
    return ""



import pytest

from .._engines import get_engine_names
from .._engines import get_execution_engine
from .._engines import get_serialization_engine


def test_engine_name_discovery():
    names = get_engine_names()
    expected = {"core", "dask", "ppf", "orange"}
    assert expected.issubset(names)


def test_execution_engine_discovery():
    core_engine = get_execution_engine("core")
    for name in (None, "none", "None"):
        engine = get_execution_engine(name)
        assert engine.__class__ is core_engine.__class__

    for engine in ("ppf", "dask", "orange"):
        _ = get_execution_engine(engine)

    with pytest.raises(RuntimeError):
        _ = get_execution_engine("__wrong_value__")


def test_serialization_engine_discovery():
    orange_engine = get_execution_engine("orange")
    engine, representation = get_serialization_engine("test.ows")
    assert engine.__class__ is orange_engine.__class__
    assert representation == "ows"

    core_engine = get_execution_engine("core")
    engine, representation = get_serialization_engine("test.json")
    assert engine.__class__ is core_engine.__class__
    assert representation == "json"

    core_engine = get_execution_engine("core")
    engine, representation = get_serialization_engine("{}")
    assert engine.__class__ is core_engine.__class__
    assert representation == "json_string"



import asyncio
import sys
from typing import Generator

if sys.platform.startswith("win"):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

import pytest
import testbook
from testbook.client import TestbookNotebookClient

from .notebooks import notebook_path

_NOTEBOOK_OUTPUTS = {"running_workflows.ipynb": "result = 16"}


@pytest.mark.parametrize("name", _NOTEBOOK_OUTPUTS)
def test_notebooks(name):
    expected_output = _NOTEBOOK_OUTPUTS[name]
    with notebook_path(name) as filename:
        decorator = testbook.testbook(filename)

        output_found = False

        def verify_notebook(tb: TestbookNotebookClient) -> None:
            tb.execute()

            nonlocal output_found
            for output in _iter_cell_outputs(tb):
                print(output)
                if expected_output in output:
                    output_found = True
                    break

        decorator(verify_notebook)()

        assert output_found


def _iter_cell_outputs(tb: TestbookNotebookClient) -> Generator[str, None, None]:
    for cell in tb.nb.cells:
        output_nodes = cell.get("outputs", [])
        for output_node in output_nodes:
            if output_node["output_type"] == "stream":
                yield output_node["text"]



# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


import importlib.metadata

release = importlib.metadata.version("ewoks")

project = "ewoks"
version = ".".join(release.split(".")[:2])
copyright = "2021-present, ESRF"
author = "ESRF"
docstitle = f"{project} {version}"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinxarg.ext",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
    "nbsphinx",
    "nbsphinx_link",
    "sphinx_copybutton",
]
templates_path = ["_templates"]
exclude_patterns = ["build", "**.ipynb_checkpoints"]

always_document_param_types = True

autosummary_generate = True
autodoc_default_flags = [
    "members",
    "undoc-members",
    "show-inheritance",
]

copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_title = docstitle
# html_logo = "_static/logo.png"
html_static_path = ["_static"]
html_template_path = ["_templates"]
html_css_files = ["custom.css"]

html_theme_options = {
    "icon_links": [
        {
            "name": "gitlab",
            "url": "https://gitlab.esrf.fr/workflow/ewoks/ewoks",
            "icon": "fa-brands fa-gitlab",
        },
        {
            "name": "pypi",
            "url": "https://pypi.org/project/ewoks",
            "icon": "fa-brands fa-python",
        },
    ],
    "logo": {
        "text": docstitle,
    },
    "footer_start": ["copyright"],
    "footer_end": ["footer_end"],
}



from argparse import Namespace
from typing import List

from ewoksutils.cli_utils import cli_arguments
from ewoksutils.cli_utils import cli_log_utils
from ewoksutils.cli_utils import cli_parse
from ewoksutils.cli_utils.cli_spec import CLIArg

from .._engines import get_graph_representations
from .cli_parse import parse_destinations


def convert_arguments(
    shell: bool = False, default_log_level: str = "warning"
) -> List[CLIArg]:
    if shell:
        args_list = cli_log_utils.log_arguments(default_log_level=default_log_level)
    else:
        args_list = []

    args_list += cli_arguments.workflow_arguments("convert")
    args_list += cli_arguments.ewoks_inputs_arguments()
    args_list += [
        CLIArg(
            "destination",
            [],
            type=str,
            help="Destination of the conversion (e.g., JSON filename).",
        ),
        CLIArg(
            "source_representation",
            ["--src-format"],
            type=str.lower,
            choices=get_graph_representations(),
            help="Source format.",
        ),
        CLIArg(
            "destination_representation",
            ["--dst-format"],
            type=str.lower,
            choices=get_graph_representations(),
            help="Destination format.",
        ),
        CLIArg(
            "load_options",
            ["-o", "--load-option"],
            action="append",
            metavar="OPTION=VALUE",
            help="Load options.",
        ),
        CLIArg(
            "save_options",
            ["-s", "--save-option"],
            action="append",
            metavar="OPTION=VALUE",
            help="Save options.",
        ),
        CLIArg(
            "exclude_requirements",
            ["--exclude-requirements"],
            action="store_true",
            help="Do not include the packages of the current Python environment as requirements in the destination workflow.",
        ),
    ]
    return args_list


def parse_convert_arguments(cli_args: Namespace, shell: bool = False) -> None:
    if shell:
        cli_log_utils.parse_log_arguments(cli_args)
    cli_args.workflows, cli_args.graphs = cli_parse.parse_workflows(cli_args)
    cli_args.destinations = parse_destinations(cli_args)

    load_options = dict(cli_parse.parse_option(item) for item in cli_args.load_options)
    if cli_args.source_representation:
        load_options["representation"] = cli_args.source_representation
    if cli_args.test:
        load_options["representation"] = "test_core"
    if cli_args.root_module:
        load_options["root_module"] = cli_args.root_module
    if cli_args.root_dir:
        load_options["root_dir"] = cli_args.root_dir

    save_options = dict(cli_parse.parse_option(item) for item in cli_args.save_options)
    if cli_args.destination_representation:
        save_options["representation"] = cli_args.destination_representation

    convert_options = {
        "save_options": save_options,
        "load_options": load_options,
        "inputs": cli_parse.parse_ewoks_inputs_parameters(cli_args),
    }
    if cli_args.exclude_requirements:
        convert_options["save_requirements"] = False
    cli_args.convert_options = convert_options



from unittest.mock import MagicMock

import pytest


@pytest.fixture
def mock_icat_client(monkeypatch):
    mock_defaults = MagicMock()
    mock_defaults.METADATA_BROKERS = ["mock://broker"]

    # Create a mock for IcatClient
    mock_client = MagicMock()
    mock_client_instance = mock_client.return_value
    mock_client_instance.store_processed_data.return_value = None

    # Patch the imports in your module (replace 'your_module' with actual module name)
    monkeypatch.setattr("ewoks.bindings.icat_defaults", mock_defaults)
    monkeypatch.setattr("ewoks.bindings.IcatClient", mock_client)

    return mock_client



from ewokscore import Task  # noqa: F401
from ewokscore.model import BaseInputModel  # noqa: F401

from .bindings import convert_graph  # noqa: F401
from .bindings import execute_graph  # noqa: F401
from .bindings import load_graph  # noqa: F401
from .bindings import save_graph  # noqa: F401
from .bindings import show_graph  # noqa: F401
from .bindings import submit_graph  # noqa: F401



import logging
from argparse import Namespace
from typing import List

from ewoksutils.cli_utils import cli_arguments
from ewoksutils.cli_utils import cli_log_utils
from ewoksutils.cli_utils import cli_parse
from ewoksutils.cli_utils.cli_spec import CLIArg

logger = logging.getLogger(__name__)


def install_arguments(
    shell: bool = False, default_log_level: str = "info"
) -> List[CLIArg]:
    if shell:
        args_list = cli_log_utils.log_arguments(default_log_level=default_log_level)
    else:
        args_list = []

    args_list += cli_arguments.workflow_arguments("install")
    args_list += [
        CLIArg(
            "yes",
            ["--yes"],
            action="store_true",
            help="Automatically accept installation prompts.",
        ),
        CLIArg(
            "python",
            ["-p", "--python"],
            type=str,
            help="Python interpreter of the environment where the packages should be installed. Default: current environment Python.",
        ),
    ]
    return args_list


def parse_install_arguments(cli_args: Namespace, shell: bool = False) -> None:
    if shell:
        cli_log_utils.parse_log_arguments(cli_args)
    cli_args.workflows, cli_args.graphs = cli_parse.parse_workflows(cli_args)
